#ifdef _WIN32
#define WIN32NATIVE
//#define HAVE_TERMIOS
#endif

#if defined(__linux__) || defined(WIN32NATIVE)
#define HAVE_ARGS
#define USE_INTERNAL_FMUL
#define EG_DEBUG
#elif defined(__AVR_ARCH__)
#if (__AVR_ARCH__==2)
#define USE_INTERNAL_FMUL
//#define NO_LFO
//#define NO_EG
#endif
#endif

#ifdef __linux__
#include <stdio.h>
#include <stdlib.h>

typedef unsigned char uint8_t;
typedef unsigned short uint16_t;
typedef unsigned int uint32_t;

#define PSTR(str)	(str)
#define prog_char	char
#define prog_uint16_t	uint16_t
#define __LPM(p)	(*(p))
#define __LPM_word(p)	(*(p))
#define strcmp_P	strcmp
#define strcpy_P	strcpy
#define sei()
#define cpuclk_updown(v)

#define O_BINARY	0

#elif defined(WIN32NATIVE)

#include <windows.h>
#include <io.h>
#include <stdio.h>
#include <stdlib.h>

typedef unsigned char uint8_t;
typedef unsigned short uint16_t;
typedef unsigned int uint32_t;

#define PSTR(str)	(str)
#define prog_char	char
#define prog_uint16_t	uint16_t
#define __LPM(p)	(*(p))
#define __LPM_word(p)	(*(p))
#define strcmp_P	strcmp
#define strcpy_P	strcpy
#define sei()
#define cpuclk_updown(v)

#elif defined(__AVR_ARCH__)

/*
 *               ATtiny45/85 Develop Version 
 *
 *                ~RESET  1      8 VCC
 *        +--------- PB3  2      7 SCK/SCL --------- I2C SCL / ISP
 *        | +-- OC1B/PB4  3      6 PB1/OC1A ---+----      ISP
 *        | |        GND  4      5 MOSI/SDA ---|---- I2C SDA / ISP
 *        | |                                  |
 *      +--------------------------------------+
 *      | | |
 *      | | +-- L1 --+--- C2 -- R1 ---> Headhone Out (L)
 *      | |          |
 *      o |          C1
 *  SW  o |          |
 *      | |       ---+---
 *      | |  
 *      +------ L2 --+--- C4 -- R2 ---> Headhone Out (R)  (option)
 *        |          |
 *        |          C3
 *        |          |
 *        |       ---+---
 *        |
 *        +---- R3 ---+----> CPU Usage (option2)
 *                    |
 *                    C5
 *                    |
 *                 ---+---
 *
 * L1,L2  47uH
 * C1,C3  10u
 * C2,C4  100u
 * C5     10u
 * R1,R2  75 - 150
 * R3     4.7k
 *
 *  FUSE Low: 62 -> 61 (16MHz)
 *
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <avr/wdt.h>
#include <util/delay.h>
#include <stdlib.h>
#include <string.h>

#define bit_set(port,bit)	{ port |= (1<<(bit));}
#define bit_clear(port,bit)	{ port &= ~(1<<(bit));}

#endif

#include "snd.h"
#include "i2cdev.h"

uint16_t sndbuf_underflow;

#if defined(__AVR_ARCH__)

#if defined(__AVR_ATtiny25__) || defined(__AVR_ATtiny45__)
#define PWM_PORT	PORTB
#define PWM_DDR		DDRB
#define PWM_BIT_L	PB4
#define PWM_BIT_R	PB1
#define CPU_USAGE
#define CPU_PORT	PORTB
#define CPU_DDR		DDRB
#define CPU_BIT		PB3
#endif

#define SNDBUF_SIZE	32

uint8_t sndbuf[SNDBUF_SIZE];
uint8_t sndbuf_in;
uint8_t sndbuf_out;

SIGNAL(SIG_OUTPUT_COMPARE0A) {
	int8_t v;
#ifdef CPU_USAGE_L2
	uint8_t x = 0;
	if (bit_is_clear(CPU_PORT, CPU_BIT)) {
		bit_set(CPU_PORT, CPU_BIT);
		x = 1;
	}
#endif
	if (sndbuf_out == sndbuf_in) {
		sndbuf_underflow++;
		return;
	}
	v = sndbuf[sndbuf_out];
	OCR1B = v;
	sndbuf_out = (sndbuf_out+1) & (SNDBUF_SIZE-1);
#ifdef CPU_USAGE_L2
	if (x) {
		bit_clear(CPU_PORT, CPU_BIT);
	}
#endif
}

void sndout(int8_t v) {
	uint8_t next;
	
	next = (sndbuf_in+1) & (SNDBUF_SIZE-1);
	if (next == sndbuf_out) {
		return;
	}
	sndbuf[sndbuf_in] = v + 127;
	sndbuf_in = next;
}

int can_sndout() {
	uint8_t next;
	next = (sndbuf_in+1) & (SNDBUF_SIZE-1);
	return !(next == sndbuf_out);
}

void sndout_init() {
	bit_set(PWM_DDR, PWM_BIT_L);
	//bit_set(PWM_DDR, PWM_BIT_R);

#if defined(__AVR_ATtiny45__) || defined(__AVR_ATtiny85__)
#if 1
	/* TIMER0 (SAMPLING RATE) setup */
	OCR0A = 250 - 1;
	TCCR0A = (2<<WGM00); // CTC 
	TCCR0B = 2; // start @ 1/8 
	TIMSK = (1<<OCIE0A);
#else
	/* TIMER0 (SAMPLING RATE) setup */
	//TCCR0B = 2; // start @ 1/8 : 7812 Hz
	TIMSK = (1<<TOIE0);
#endif
#if 1
	/* TIMER1 (PWM) setup */
	bit_set(PLLCSR, PLLE);
	while (bit_is_clear(PLLCSR, PLOCK))
		;
	PLLCSR |= (1<<PCKE);

	OCR1B = 127;
	GTCCR = (2<<COM1B0) | (1<<PWM1B);
	OCR1A = 127;
	TCCR1 = (2<<COM1A0) | (1<<PWM1A);
	TCCR1 |= 1; // start @ 1/1 
#endif

#endif
}

void cpuclk_updown(int8_t updown) {
	if (updown > 0) {
		OSCCAL++;
	} else if (updown < 0) {
		OSCCAL--;
	}
}

#endif // AVR

#define WAVE_SIN  0
#define WAVE_TRI  1
#define WAVE_SAW  2

prog_char sintab[64] = {
    0,    3,    6,    9,   12,   15,   18,   21,
   24,   27,   30,   33,   36,   39,   42,   45,
   48,   51,   54,   57,   59,   62,   65,   67,
   70,   73,   75,   78,   80,   82,   85,   87,
   89,   91,   94,   96,   98,  100,  102,  103,
  105,  107,  108,  110,  112,  113,  114,  116,
  117,  118,  119,  120,  121,  122,  123,  123,
  124,  125,  125,  126,  126,  126,  126,  126,
};

// warning: max RAMPLE_RATE = 32K
prog_uint16_t note_tab[12] = {
	137167254/2/ SAMPLE_RATE, // C6 = 96
	145323630/2/ SAMPLE_RATE,
	153965009/2/ SAMPLE_RATE,
	163120231/2/ SAMPLE_RATE,
	172819849/2/ SAMPLE_RATE,
	183096236/2/ SAMPLE_RATE,
	193983687/2/ SAMPLE_RATE,
	205518539/2/ SAMPLE_RATE,
	217739288/2/ SAMPLE_RATE,
	230686720/2/ SAMPLE_RATE, // A6 1760 Hz
	244404044/2/ SAMPLE_RATE,
	258937042/2/ SAMPLE_RATE,
};

struct LFO {
	uint16_t wave_pos;
	int16_t wave_inc;  // 2601 = 20Hz (20 * 255 * 16/SAMPLE_RATE)
	uint8_t intensity; // 255 = 6.225%
} lfo;

struct VCO {
	uint16_t wave_pos;
	int16_t wave_inc;
	int8_t lfo_inc;
	uint8_t wave_type;
} vco;

#define EG_NOTEOFF	0
#define EG_NOTEON	1
#define EG_ATTACK 	2
#define EG_DECAY	3
#define EG_SUSTAIN	4
#define EG_RELEASE	5

/* EG tick = 8000 / 16 = 500 Hz */
struct EG {
	uint8_t state;
	uint8_t out;
	uint8_t velocity;
	uint16_t attack;
	uint16_t decay;
	uint16_t sustain_time; // 500Hz : max 131 sec
	uint16_t release;
	uint16_t sustain_lvl; // 0xffff : 100 %
	uint16_t sustain_cnt; // work
	uint16_t intval; // work
} eg;



#ifdef USE_INTERNAL_FMUL

static inline uint16_t fmul16(uint16_t a, uint16_t b) {
	uint8_t i;
	uint32_t __c , __a;
	__c = 0;
	__a = a;
	
	for (i=0; i<16; i++) {
		if (b & 1) {
			__c += __a;
		}
		b>>=1;
		__a <<= 1;
	}
	return __c>>16;
}

static inline int16_t fmulsu16(int16_t a, uint16_t b) {
	uint8_t i;
	int32_t __c , __a;
	__c = 0;
	__a = a;
	
	for (i=0; i<16; i++) {
		if (b & 1) {
			__c += __a;
		}
		b>>=1;
		__a <<= 1;
	}
	return __c>>16;
}

static inline uint16_t fmul16_8(uint16_t a, uint8_t b) {
	uint8_t i;
	uint32_t __c , __a;
	__c = 0;
	__a = a;
	
	for (i=0; i<8; i++) {
		if (b & 1) {
			__c += __a;
		}
		b>>=1;
		__a <<= 1;
	}
	return __c>>8;
}

static inline int16_t fmulsu16_8(int16_t a, uint8_t b) {
	uint8_t i;
	int32_t __c , __a;
	__c = 0;
	__a = a;
	
	for (i=0; i<8; i++) {
		if (b & 1) {
			__c += __a;
		}
		b>>=1;
		__a <<= 1;
	}
	return __c>>8;
}

static inline int8_t fmulsu8_8(int8_t a, uint8_t b) {
	uint8_t i;
	int16_t __c , __a;
	__c = 0;
	__a = a;
	
	for (i=0; i<8; i++) {
		if (b & 1) {
			__c += __a;
		}
		b>>=1;
		__a <<= 1;
	}
	return __c>>8;
}
#else

static inline uint16_t fmul16(uint16_t a, uint16_t b) {
	uint32_t v;
	v = (uint32_t)a * b;
	return (uint16_t)(v >> 16);
}

static inline int16_t fmulsu16(int16_t a, uint16_t b) {
	int32_t v;
	v = (int32_t)a * b;
	return (int16_t)(v >> 16);
}

static inline uint16_t fmul16_8(uint16_t a, uint8_t b) {
	uint32_t v;
	v = (uint32_t)a * b;
	return (uint16_t)(v >> 8);
}

static inline int16_t fmulsu16_8(int16_t a, uint8_t b) {
	int32_t v;
	v = (int32_t)a * b;
	return (int16_t)(v >> 8);
}

static inline int8_t fmulsu8_8(int8_t a, uint8_t b) {
	int16_t v;
	v = (int16_t)a * b;
	return (int8_t)(v >> 8);
}
#endif


static void note_on(uint8_t note,uint8_t velocity) {
	uint16_t v;
	uint8_t shift = 0;
	if (note >= 108) return; 
	while (note < 96) {
		note += 12;
		shift++;
	}
	v = __LPM_word(note_tab+note-96);
	v >>= shift;
	vco.wave_inc = v;
	eg.velocity = velocity << 1;
	eg.state = EG_NOTEON;
}

static void note_off() {
#ifndef NO_EG
	eg.state = EG_NOTEOFF;
#else
	vco.wave_inc = 0;
#endif
}

static inline void do_eg() {
	uint8_t state = eg.state;
	if (state == EG_NOTEOFF) {
		eg.intval = (0 << 8);
	}
	if (state == EG_NOTEON) {
		eg.sustain_cnt = 0;
		if (eg.attack == 0) {
			eg.intval = eg.sustain_lvl;
			eg.state = state = EG_SUSTAIN;
#ifdef EG_DEBUG
fprintf(stderr,"EG_SUSTAIN\n");
#endif
		} else {
			eg.intval = 0;
			eg.state = state = EG_ATTACK;
#ifdef EG_DEBUG
fprintf(stderr,"EG_ATTACK %04x\n",eg.attack);
#endif
		}
	}
	if (state == EG_ATTACK) {
		uint16_t next = eg.intval + eg.attack;
		if (next < eg.intval) { // overflow
			eg.intval = (255 << 8);
			if (eg.sustain_lvl == 0) {
#ifdef EG_DEBUG
fprintf(stderr,"EG_RELEASE\n");
#endif
				eg.state = EG_RELEASE;
			} else {
#ifdef EG_DEBUG
fprintf(stderr,"EG_DECAY\n");
#endif
				eg.state = EG_DECAY;
			}
		} else {
			eg.intval = next;
		}
	} else if (state == EG_DECAY) {
		eg.intval = fmul16(eg.intval, eg.decay);
		if (eg.intval <= eg.sustain_lvl) {
			eg.intval = eg.sustain_lvl;
			if (eg.sustain_time == 0) {
#ifdef EG_DEBUG
fprintf(stderr,"EG_RELEASE\n");
#endif
				eg.state = EG_RELEASE;
			} else {
#ifdef EG_DEBUG
fprintf(stderr,"EG_SUSTAIN\n");
#endif
				eg.state = EG_SUSTAIN;
			}
		}
	} else if (state == EG_SUSTAIN) {
		eg.sustain_cnt ++;
		if (eg.sustain_cnt == eg.sustain_time) {
			eg.state = EG_RELEASE;
#ifdef EG_DEBUG
fprintf(stderr,"EG_RELEASE\n");
#endif
		}
	} else if (state == EG_RELEASE) {
		eg.intval = fmul16(eg.intval, eg.release);
		if (eg.intval < (1 << 8)) {
			eg.intval = 0;
			eg.state = EG_NOTEOFF;
#ifdef EG_DEBUG
fprintf(stderr,"EG_NOTEOFF\n");
#endif
		}
	}
	eg.out = fmul16_8(eg.intval, eg.velocity) >> 8;
}

static inline void do_lfo() {
	uint8_t pos;
	int8_t rev = 0;
	int8_t v;
	int16_t v2;
	lfo.wave_pos += lfo.wave_inc;
	pos = lfo.wave_pos >> 8;
	if (pos >= 128) {
		rev = 1;
		pos = 255 - pos;
	}
	if (pos >= 64) {
		pos = 127 - pos;
	}
	v = __LPM(sintab+pos);
	v2 = fmul16_8(vco.wave_inc, lfo.intensity);
	if (rev) v2 = -v2;
	vco.lfo_inc = fmulsu16_8(v2, v)>>4;
}

static void generate_snd() {
	uint8_t pos;
	int8_t rev = 0;
	int8_t v;
	static uint8_t cnt;

	vco.wave_pos += vco.wave_inc + vco.lfo_inc;
	pos = vco.wave_pos >> 8;
	if (pos >= 128) {
		rev = 1;
		pos = 255 - pos;
	}
	if (vco.wave_type == WAVE_SAW) {
		v = 127 - pos;
	} else {
		if (pos >= 64) {
			pos = 127 - pos;
		}
		if (vco.wave_type == WAVE_TRI) {
			v = pos*2;
		} else 
			v = __LPM(sintab+pos);
	}
	if (rev) v = -v;
	cnt++;
	if (cnt >= 16) {
#ifndef NO_LFO
		do_lfo();
#endif
		cnt = 0;
	} else if (cnt == 8) {
#ifndef NO_EG
		do_eg();
#endif
	}
#ifndef NO_EG
	v = fmulsu8_8(v, eg.out);
#else
	//v = fmulsu8_8(v, 255);
	//v = fmulsu16(v<<8, 255<<8) >> 8;
#endif
	sndout(v);
}

#define CMD_NOTEON	0x90 
#define CMD_NOTEOFF	0x80
#define CMD_SETPARAM	0x70

#define PARAM_LFO_FREQ		0
#define PARAM_LFO_INTENSITY	1
#define PARAM_VCO_WAVE		2
#define PARAM_EG_ATTACK		3
#define PARAM_EG_DECAY		4
#define PARAM_EG_SUSTAIN_LVL	5
#define PARAM_EG_SUSTAIN_TIME	6
#define PARAM_EG_RELEASE	7

#define PARAM_SYS_CPUCLK	128

static void command_proc() {
	uint8_t cmd,p1,p2,p3; 
	cmd = i2cdev_getc() & 0xf0;
	p1 = i2cdev_getc();
	p2 = i2cdev_getc();
	p3 = i2cdev_getc();
//printf("command %02x %02x %02x %02x\n",cmd,p1,p2,p3);
	if (cmd == CMD_NOTEON) {
		note_on(p1,p2);
	} else if (cmd == CMD_NOTEOFF) {
		note_off();
	} else if (cmd == CMD_SETPARAM) {
		switch(p1) {
		case PARAM_LFO_FREQ:
			lfo.wave_inc = ((int16_t)p3 << 8) | p2;  break;
		case PARAM_LFO_INTENSITY:
			lfo.intensity = p2; break;
		case PARAM_VCO_WAVE:
			vco.wave_type = p2; break;
		case PARAM_EG_ATTACK: 
			eg.attack = ((int16_t)p3 << 8) | p2;  break;
		case PARAM_EG_DECAY:
			eg.decay =  ((int16_t)p3 << 8) | p2;  break;
		case PARAM_EG_SUSTAIN_LVL: 
			eg.sustain_lvl = ((int16_t)p3 << 8) | p2;  break; 
		case PARAM_EG_SUSTAIN_TIME:
			eg.sustain_time = ((int16_t)p3 << 8) | p2;  break; 
		case PARAM_EG_RELEASE: 
			eg.release = ((int16_t)p3 << 8) | p2;  break; 
		case PARAM_SYS_CPUCLK: 
			cpuclk_updown(p2); break;
		}
	}
}

#ifdef HAVE_ARGS
int main(int argc,char *argv[]) {
	uint8_t c;
	char *arg;
#else // AVR
int main() {

	/* set system_clock pre scaler 1/8 => 1/1 */
	CLKPR = (1<<CLKPCE);
	CLKPR = 0;
#endif

#ifdef NO_EG
	lfo.intensity = 255;
	lfo.wave_inc = 262; // 2 Hz
	//lfo.wave_inc = 131; // 1 Hz
	eg.sustain_lvl = (255<<8);
	eg.sustain_time = 500*4; // 4 sec
#else // NO_EG

#if 0 // for demo
	eg.attack = 64;
	eg.sustain_lvl = (128<<8);
	eg.sustain_time = (12<<8);
	eg.decay = 0xfff0;
	eg.release = 0xff80;
#endif
#if 0 // SUSTAIN and RELEASE test
	eg.sustain_lvl = (127<<8);
	eg.sustain_time = 1500; // 3 sec
	eg.release = 0xff80;
#endif
#if 1 // normal
	eg.attack = 64<<8;
	eg.decay = 0xff80;
	eg.sustain_lvl = (12<<8);
	eg.sustain_time = 0;
	eg.release = 0xf000;
	lfo.intensity = 255;
	lfo.wave_inc = 262; // 2 Hz
#endif

#endif // NO_EG

	i2cdev_init(0x20);
	sndout_init();
#ifdef NO_EG
	note_on(36+24,127); // for test
#endif

	sei();

#ifdef CPU_USAGE
	bit_set(CPU_DDR, CPU_BIT);
#endif
	for (;;) {
		if (can_sndout()) {
#ifdef CPU_USAGE
			bit_set(CPU_PORT, CPU_BIT);
#endif
			generate_snd();
#ifdef CPU_USAGE
			bit_clear(CPU_PORT, CPU_BIT);
#endif
			continue;
		}
		if (i2cdev_can_getc(1)) {
#ifdef CPU_USAGE
			bit_set(CPU_PORT, CPU_BIT);
#endif
			command_proc();
			i2cdev_putc(sndbuf_underflow>>8);
			i2cdev_putc(sndbuf_underflow);
			i2cdev_flush();
#ifdef CPU_USAGE
			bit_clear(CPU_PORT, CPU_BIT);
#endif
		}
	}
}
