#ifdef _WIN32
#define WIN32NATIVE
//#define HAVE_TERMIOS
#endif

#ifdef __linux__
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <errno.h>
#include <termios.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <linux/soundcard.h>

typedef unsigned char uint8_t;
typedef unsigned short uint16_t;
typedef unsigned int uint32_t;

#define PSTR(str)	(str)
#define prog_char	char
#define prog_uint16_t	uint16_t
#define __LPM(p)	(*(p))
#define __LPM_word(p)	(*(p))
#define strcmp_P	strcmp
#define strcpy_P	strcpy

#define O_BINARY	0

#elif defined(WIN32NATIVE)

#include <windows.h>
#include <io.h>
#ifdef HAVE_TERMIOS
#include <termios.h>
#else
#include <wincon.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <errno.h>
#include <fcntl.h>

typedef unsigned char uint8_t;
typedef unsigned short uint16_t;
typedef unsigned int uint32_t;

#define PSTR(str)	(str)
#define prog_char	char
#define prog_uint16_t	uint16_t
#define __LPM(p)	(*(p))
#define __LPM_word(p)	(*(p))
#define strcmp_P	strcmp
#define strcpy_P	strcpy

#endif

#include "snd.h"

#ifdef __linux__
int sndout_fd;
int audio_buf_size;

void sndout_init() {
	int arg,r;
	struct audio_buf_info info;

	sndout_fd = open("/dev/dsp",O_RDWR);
	if (sndout_fd < 0) {
		perror("open");
		exit(1);
	}
	arg = 16;
	r = ioctl(sndout_fd, SOUND_PCM_WRITE_BITS, &arg);
	if (r) {
		perror("SOUND_PCM_WRITE_BITS ioctl failed");
		exit(1);
	}
	arg = 2;
	r = ioctl(sndout_fd, SOUND_PCM_WRITE_CHANNELS, &arg);
	if (r) {
		perror("SOUND_PCM_WRITE_CHANNELS ioctl failed");
		exit(1);
	}
	arg = SAMPLE_RATE;
	r = ioctl(sndout_fd, SOUND_PCM_WRITE_RATE, &arg);
	if (r) {
		perror("SOUND_PCM_WRITE_RATE ioctl failed");
		exit(1);
	}
#if 0
	arg = 0;
	r = ioctl(sndout_fd, SOUND_PCM_READ_RATE, &arg);
	printf("RATE = %d\n",arg);
	arg = 0;
	r = ioctl(sndout_fd, SOUND_PCM_GETFMTS, &arg);
	printf("FORMAT = %d\n",arg);
#endif
	arg = 0x0008000c;
	r = ioctl(sndout_fd, SOUND_PCM_SETFRAGMENT, &arg);
	if (r) {
		perror("SOUND_PCM_SETFRAGMENT");
		exit(1);
	}

	r = ioctl(sndout_fd, SOUND_PCM_GETOSPACE, &info);
	fprintf(stderr,"r %d fragstotal %d fragsize %d bytes %d\n",
		r,info.fragstotal, info.fragsize, info.bytes);
	if (r) {
		perror("SOUND_PCM_GETOSPACE");
		exit(1);
	}
	audio_buf_size = info.bytes;
}

int can_sndout() {
	int r;
	struct audio_buf_info info;
	r = ioctl(sndout_fd, SOUND_PCM_GETOSPACE, &info);
	if (r) {
		perror("SOUND_PCM_GETOSPACE");
		exit(1);
	}
	if (audio_buf_size - info.bytes >= info.fragsize * 2) {
		return 0;
	}
	return 1;
}

void sndout(int8_t d) {
	int16_t buf[2];
#if 1
	static int x;
	if (!isatty(1))
	    printf("%5d %5d\n",x++, d);
#endif
	buf[0] = (int16_t)d << 8;
	buf[1] = (int16_t)d << 8;
	write(sndout_fd, buf, 4);
}

#elif defined(_WIN32)
#include <mmsystem.h>

#define WAVEBUF_SIZE	4096
#define NR_WAVEHDR	8

HWAVEOUT HWaveOut = NULL;
uint8_t *WaveBuf;
WAVEHDR WaveHdr[NR_WAVEHDR];
int cur_wavehdr = 0;
int cur_wavebuf_len = 0;
WAVEFORMATEX waveFmt;

void sndout_init() {
	int r,i;
	waveFmt.wFormatTag = WAVE_FORMAT_PCM;
	waveFmt.nChannels = 2;
	waveFmt.nSamplesPerSec = SAMPLE_RATE;
	waveFmt.wBitsPerSample = 16;
	waveFmt.nBlockAlign = waveFmt.nChannels * waveFmt.wBitsPerSample/8;
	waveFmt.nAvgBytesPerSec = SAMPLE_RATE * waveFmt.nBlockAlign;
	waveFmt.cbSize = (WORD)0;

	WaveBuf = GlobalAlloc(GPTR, NR_WAVEHDR * WAVEBUF_SIZE);
	memset(WaveHdr, 0, sizeof(WAVEHDR) * NR_WAVEHDR);
	for (i=0; i<NR_WAVEHDR; i++) {
		WaveHdr[i].lpData = WaveBuf+ i * WAVEBUF_SIZE;
		WaveHdr[i].dwBufferLength = WAVEBUF_SIZE;
	}

	r = waveOutOpen(&HWaveOut, WAVE_MAPPER, &waveFmt, 0, 0, CALLBACK_NULL);
	if (r != MMSYSERR_NOERROR) {
		fprintf(stderr,"waveOutOpen Error\n");
		exit(1);
	}
	for (i=0; i<NR_WAVEHDR; i++) {
		waveOutPrepareHeader(HWaveOut,WaveHdr + i, sizeof(WAVEHDR));
	}
}

void sndout(int8_t d) {
	int next;

	next = cur_wavehdr + 1;
	if (next >= NR_WAVEHDR) next -= NR_WAVEHDR;
	if (cur_wavebuf_len >= WAVEBUF_SIZE) {
//fprintf(stderr,"start waveOutWrite\n");
//fflush(stderr);
		while (!(WaveHdr[next].dwFlags & WHDR_DONE)
		       && (WaveHdr[next].dwFlags != WHDR_PREPARED)) {
			fprintf(stderr,"wave buffer overflow\n");
			exit(1);
		}
//fprintf(stderr,"waveOutWrite #%d\n",cur_wavehdr);
//fflush(stderr);
		waveOutWrite(HWaveOut, WaveHdr + cur_wavehdr, sizeof(WAVEHDR));
		cur_wavehdr = next;
		cur_wavebuf_len = 0;
	}
	WaveHdr[cur_wavehdr].lpData[cur_wavebuf_len++] = 0;
	WaveHdr[cur_wavehdr].lpData[cur_wavebuf_len++] = d;
	WaveHdr[cur_wavehdr].lpData[cur_wavebuf_len++] = 0;
	WaveHdr[cur_wavehdr].lpData[cur_wavebuf_len++] = d;
}

int can_sndout() {
	int prev;

	prev = cur_wavehdr - 5;
	if (prev < 0) prev += NR_WAVEHDR;
	if  (WaveHdr[prev].dwFlags == WHDR_PREPARED) return 1;
	if  (WaveHdr[prev].dwFlags & WHDR_DONE) return 1;
	return 0;
}

#endif
