#ifdef _WIN32
#define WIN32NATIVE
//#define HAVE_TERMIOS
#endif

#ifdef __linux__
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <errno.h>
#include <termios.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <linux/soundcard.h>

typedef unsigned char uint8_t;
typedef unsigned short uint16_t;
typedef unsigned int uint32_t;

#define PSTR(str)	(str)
#define prog_char	char
#define prog_uint16_t	uint16_t
#define __LPM(p)	(*(p))
#define __LPM_word(p)	(*(p))
#define strcmp_P	strcmp
#define strcpy_P	strcpy

#define O_BINARY	0

#elif defined(WIN32NATIVE)

#include <windows.h>
#include <io.h>
#ifdef HAVE_TERMIOS
#include <termios.h>
#else
#include <wincon.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <errno.h>
#include <fcntl.h>

typedef unsigned char uint8_t;
typedef unsigned short uint16_t;
typedef unsigned int uint32_t;

#define PSTR(str)	(str)
#define prog_char	char
#define prog_uint16_t	uint16_t
#define __LPM(p)	(*(p))
#define __LPM_word(p)	(*(p))
#define strcmp_P	strcmp
#define strcpy_P	strcpy

#endif

#if defined(__linux__) || defined(WIN32NATIVE)

#if defined(WIN32NATIVE) && !defined(HAVE_TERMIOS)
static HANDLE conin;
static DWORD conin_mode_sv;
#else
struct termios term_sv;
#endif

void lin_term() {
#if defined(WIN32NATIVE) && !defined(HAVE_TERMIOS)
	if (conin_mode_sv != 0) 
		SetConsoleMode(conin, conin_mode_sv);
#else
	tcsetattr(0, TCSANOW, &term_sv);
#endif
}

void lin_init() {
#if defined(WIN32NATIVE) && !defined(HAVE_TERMIOS)
	DWORD conin_mode = 0;
	int r;

	conin = GetStdHandle(STD_INPUT_HANDLE);
	r = GetConsoleMode(conin, &conin_mode);
	if (r) {
		conin_mode_sv = conin_mode;
		conin_mode &= ~(ENABLE_LINE_INPUT | ENABLE_ECHO_INPUT);
		r = SetConsoleMode(conin, conin_mode);
	} else {
		fprintf(stderr, "run on the console(command prompt)\n");
		exit(1);
	}
#else
	struct termios term;
	tcgetattr(0, &term);
	term_sv = term;
	term.c_lflag &= ~(ECHO | ICANON);
	term.c_iflag &= ~(ICRNL);
	tcsetattr(0, TCSANOW, &term);
#endif
	atexit(lin_term);

fprintf(stderr,"end lin_init\n");
fflush(stderr);
}

void usbcdc_putc(char data) {
	write(1, &data, 1);
}

char usbcdc_getc(void) {
	int r;
	char ret = 0;
#if defined(WIN32NATIVE) && !defined(HAVE_TERMIOS)
	char buf[3];
	DWORD nrec = 0;

	while (!ReadConsole(conin, buf, 1, &nrec, NULL) || nrec == 0)
			;
		return buf[0];
#endif
	while ((r = read(0, &ret, 1)) != 1)
		;

	return ret;
}

#if defined(WIN32NATIVE) && !defined(HAVE_TERMIOS)
static int8_t usbcdc_can_getc(int8_t size) {
	INPUT_RECORD ir;
	DWORD nrec = 0;
	if (PeekConsoleInput(conin, &ir, 1, &nrec)) {
		if ((nrec != 0) && (ir.EventType & KEY_EVENT) &&
			ir.Event.KeyEvent.bKeyDown ) {
fprintf(stderr,"Console %d %08x\n",ir.Event.KeyEvent.bKeyDown, ir.Event.KeyEvent.wVirtualKeyCode);
fflush(stderr);
			return 1;
		}
	}
	return 0;
}
#else
#include <sys/time.h>
#if defined(WIN32NATIVE) && defined(HAVE_TERMIOS)
#define select _select
#endif
static int8_t usbcdc_can_getc(int8_t size) {
	int r;
	fd_set fds;
	struct timeval to;
	to.tv_usec = 50000; // 50 ms
	to.tv_sec = 0;

	FD_ZERO(&fds);
	FD_SET(0, &fds);

	r = select(1, &fds, NULL, NULL, &to);
	if (r == 1) {
		return 1;
	}
	return 0;
}
#endif
#endif

#define CMD_NOTEON      0x90 
#define CMD_NOTEOFF     0x80
#define CMD_SETPARAM    0x70

#define PARAM_LFO_FREQ          0
#define PARAM_LFO_INTENSITY     1
#define PARAM_VCO_WAVE          2
	#define WAVE_SIN  0
	#define WAVE_TRI  1
	#define WAVE_SAW  2
#define PARAM_EG_ATTACK         3
#define PARAM_EG_DECAY          4
#define PARAM_EG_SUSTAIN_LVL    5
#define PARAM_EG_SUSTAIN_TIME   6
#define PARAM_EG_RELEASE        7

#define PARAM_SYS_CPUCLK	128

uint8_t ext_buf[16];

static void note_on(uint8_t note, uint8_t verosity) {
	ext_buf[0] = CMD_NOTEON;
	ext_buf[1] = note;
	ext_buf[2] = verosity;
	ext_buf[3] = 0;
	i2c_send(0x20, 4);
}

static void note_off(uint8_t note, uint8_t verosity) {
	ext_buf[0] = CMD_NOTEOFF;
	ext_buf[1] = note;
	ext_buf[2] = verosity;
	ext_buf[3] = 0;
	i2c_send(0x20, 4);
}

static void set_param(uint8_t kind, uint16_t val) {
	ext_buf[0] = CMD_SETPARAM;
	ext_buf[1] = kind;
	ext_buf[2] = val;
	ext_buf[3] = val>>8;
	i2c_send(0x20, 4);
}

#if 1
static void print_dev_status() {
	int8_t r;
	int i;
	r = i2c_recv(0x20, 2);
	printf("r %02d :",r);
	for (i=0; i< r; i++) {
		printf("%02x ", ext_buf[i]);
	}
	printf("\n");
}
#endif

static void change_note(uint8_t c) {
   const uint8_t note_base = 36+24; // C1
   static int oct;
   static int last_note;
    switch(c) {
    case 'z': oct --; break;
    case 'x': oct ++; break;
    case 'q': exit(0);
    case 'o': note_off(last_note, 0); break;
    case 'w': set_param(PARAM_VCO_WAVE, WAVE_SIN); break; 
    case 'e': set_param(PARAM_VCO_WAVE, WAVE_TRI); break; 
    case 'r': set_param(PARAM_VCO_WAVE, WAVE_SAW); break; 
    case 'a': note_on(note_base + 12*oct +0, 127); break;
    case 's': note_on(note_base + 12*oct +2, 127); break;
    case 'd': note_on(note_base + 12*oct +4, 127); break;
    case 'f': note_on(note_base + 12*oct +5, 127); break;
    case 'g': note_on(note_base + 12*oct +7, 127); break;
    case 'h': note_on(note_base + 12*oct +9, 127); break;
    case 'j': note_on(note_base + 12*oct +11, 127); break;
    case 'k': note_on(note_base + 12*oct +12, 127); break;
    case 'l': note_on(note_base + 12*oct +14, 127); break;
#if 1
    case 'n': print_dev_status(); break;
    case 'm': print_host_status(); break;
    case '<': set_param(PARAM_SYS_CPUCLK, -1); break; 
    case '>': set_param(PARAM_SYS_CPUCLK, 1); break; 
#endif
    }
}

char *port_name;
int main(int argc,char *argv[]) {
	uint8_t c;
	char *arg;

	if (argc > 1) {
		port_name = argv[1];
	}
	lin_init();
	init_i2c();
	for (;;) {
		c = usbcdc_getc();
		change_note(c);
	}
}
