#ifdef _WIN32
#define WIN32NATIVE
//#define HAVE_TERMIOS
#endif

#ifdef __linux__
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <errno.h>
#include <termios.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <linux/soundcard.h>

typedef unsigned char uint8_t;
typedef unsigned short uint16_t;
typedef unsigned int uint32_t;

#define PSTR(str)	(str)
#define prog_char	char
#define prog_uint16_t	uint16_t
#define __LPM(p)	(*(p))
#define __LPM_word(p)	(*(p))
#define strcmp_P	strcmp
#define strcpy_P	strcpy

#define O_BINARY	0

#elif defined(WIN32NATIVE)

#include <windows.h>
#include <io.h>
#ifdef HAVE_TERMIOS
#include <termios.h>
#else
#include <wincon.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <errno.h>
#include <fcntl.h>

typedef unsigned char uint8_t;
typedef unsigned short uint16_t;
typedef unsigned int uint32_t;

#define PSTR(str)	(str)
#define prog_char	char
#define prog_uint16_t	uint16_t
#define __LPM(p)	(*(p))
#define __LPM_word(p)	(*(p))
#define strcmp_P	strcmp
#define strcpy_P	strcpy

#else /* AVR */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <avr/wdt.h>
#include <util/delay.h>
#include <stdlib.h>
#include <string.h>

//#include "usbcdc.h"
//#include "i2c.h"

#endif

#ifndef EXT_BUFSIZE
#define EXT_BUFSIZE	16
#endif

#ifndef EINVAL
#define EINVAL		22
#endif
#ifndef ENOENT
#define ENOENT		 2
#endif
#ifndef ENOSPC
#define ENOSPC		28
#endif
#ifndef EIO
#define EIO		 5
#endif

uint8_t ext_buf[EXT_BUFSIZE];

#if defined(__linux__) || defined(WIN32NATIVE)

#if defined(WIN32NATIVE) && !defined(HAVE_TERMIOS)
static HANDLE conin;
static DWORD conin_mode_sv;
#else
struct termios term_sv;
#endif

void lin_term() {
#if defined(WIN32NATIVE) && !defined(HAVE_TERMIOS)
	if (conin_mode_sv != 0) 
		SetConsoleMode(conin, conin_mode_sv);
#else
	tcsetattr(0, TCSANOW, &term_sv);
#endif
}

void lin_init() {
#if defined(WIN32NATIVE) && !defined(HAVE_TERMIOS)
	DWORD conin_mode = 0;
	int r;

	conin = GetStdHandle(STD_INPUT_HANDLE);
	r = GetConsoleMode(conin, &conin_mode);
	if (r) {
		conin_mode_sv = conin_mode;
		conin_mode &= ~(ENABLE_LINE_INPUT | ENABLE_ECHO_INPUT);
		r = SetConsoleMode(conin, conin_mode);
	} else {
		fprintf(stderr, "run on the console(command prompt)\n");
		exit(1);
	}
#else
	struct termios term;
	tcgetattr(0, &term);
	term_sv = term;
	term.c_lflag &= ~(ECHO | ICANON);
	term.c_iflag &= ~(ICRNL);
	tcsetattr(0, TCSANOW, &term);
#endif
	atexit(lin_term);

fprintf(stderr,"end lin_init\n");
fflush(stderr);
}

void usbcdc_putc(char data) {
	write(1, &data, 1);
}

char usbcdc_getc(void) {
	int r;
	char ret = 0;
#if defined(WIN32NATIVE) && !defined(HAVE_TERMIOS)
	char buf[3];
	DWORD nrec = 0;

	while (!ReadConsole(conin, buf, 1, &nrec, NULL) || nrec == 0)
			;
		return buf[0];
#endif
	while ((r = read(0, &ret, 1)) != 1)
		;

	return ret;
}

#if defined(WIN32NATIVE) && !defined(HAVE_TERMIOS)
static int8_t usbcdc_can_getc(int8_t size) {
	INPUT_RECORD ir;
	DWORD nrec = 0;
	if (PeekConsoleInput(conin, &ir, 1, &nrec)) {
		if ((nrec != 0) && (ir.EventType & KEY_EVENT) &&
			ir.Event.KeyEvent.bKeyDown ) {
fprintf(stderr,"Console %d %08x\n",ir.Event.KeyEvent.bKeyDown, ir.Event.KeyEvent.wVirtualKeyCode);
fflush(stderr);
			return 1;
		}
	}
	return 0;
}
#else
#include <sys/time.h>
#if defined(WIN32NATIVE) && defined(HAVE_TERMIOS)
#define select _select
#endif
static int8_t usbcdc_can_getc(int8_t size) {
	int r;
	fd_set fds;
	struct timeval to;
	to.tv_usec = 50000; // 50 ms
	to.tv_sec = 0;

	FD_ZERO(&fds);
	FD_SET(0, &fds);

	r = select(1, &fds, NULL, NULL, &to);
	if (r == 1) {
		return 1;
	}
	return 0;
}
#endif

#else /* AVR */

static inline void timer_init() {
#if defined(__AVR_ATtiny861__) || defined(__AVR_ATtiny461__)
	TCCR0A = (1<<TCW0); // 16 bit mode
	TCCR0B = 1; // timer0 start @ clk/1
#else
	TCCR1B = 1; // timer1 start @ clk/1
#endif
}

#endif

/*---------------------------------------------------------------------*/

#define WAVE_SIN  0
#define WAVE_TRI  1
#define WAVE_SAW  2

#define SAMPLE_RATE	8000

#ifdef __linux__
int sndout_fd;
int audio_buf_size;

static void sndout_init() {
	int arg,r;
	struct audio_buf_info info;

	sndout_fd = open("/dev/dsp",O_RDWR);
	if (sndout_fd < 0) {
		perror("open");
		exit(1);
	}
	arg = 16;
	r = ioctl(sndout_fd, SOUND_PCM_WRITE_BITS, &arg);
	if (r) {
		perror("SOUND_PCM_WRITE_BITS ioctl failed");
		exit(1);
	}
	arg = 2;
	r = ioctl(sndout_fd, SOUND_PCM_WRITE_CHANNELS, &arg);
	if (r) {
		perror("SOUND_PCM_WRITE_CHANNELS ioctl failed");
		exit(1);
	}
	arg = SAMPLE_RATE;
	r = ioctl(sndout_fd, SOUND_PCM_WRITE_RATE, &arg);
	if (r) {
		perror("SOUND_PCM_WRITE_RATE ioctl failed");
		exit(1);
	}
	arg = 0x0008000c;
	r = ioctl(sndout_fd, SOUND_PCM_SETFRAGMENT, &arg);
	if (r) {
		perror("SOUND_PCM_SETFRAGMENT");
		exit(1);
	}

	r = ioctl(sndout_fd, SOUND_PCM_GETOSPACE, &info);
	fprintf(stderr,"r %d fragstotal %d fragsize %d bytes %d\n",
		r,info.fragstotal, info.fragsize, info.bytes);
	if (r) {
		perror("SOUND_PCM_GETOSPACE");
		exit(1);
	}
	audio_buf_size = info.bytes;
}

static int can_sndout() {
	int r;
	struct audio_buf_info info;
	r = ioctl(sndout_fd, SOUND_PCM_GETOSPACE, &info);
	if (r) {
		perror("SOUND_PCM_GETOSPACE");
		exit(1);
	}
	if (audio_buf_size - info.bytes >= info.fragsize * 2) {
		return 0;
	}
	return 1;
}

static void sndout(int8_t d) {
	int16_t buf[2];
#if 1
	static int x;
	if (!isatty(1))
	    printf("%5d %5d\n",x++, d);
#endif
	buf[0] = (int16_t)d << 8;
	buf[1] = (int16_t)d << 8;
	write(sndout_fd, buf, 4);
}

#elif defined(_WIN32)
#include <mmsystem.h>

#define WAVEBUF_SIZE	1024
#define NR_WAVEHDR	8

HWAVEOUT HWaveOut = NULL;
uint8_t *WaveBuf;
WAVEHDR WaveHdr[NR_WAVEHDR];
int cur_wavehdr = 0;
int cur_wavebuf_len = 0;
WAVEFORMATEX waveFmt;

static void sndout_init() {
	int r,i;
	waveFmt.wFormatTag = WAVE_FORMAT_PCM;
	waveFmt.nChannels = 1;
	waveFmt.nSamplesPerSec = SAMPLE_RATE;
	waveFmt.wBitsPerSample = 8;
	waveFmt.nBlockAlign = waveFmt.nChannels * waveFmt.wBitsPerSample/8;
	waveFmt.nAvgBytesPerSec = SAMPLE_RATE * waveFmt.nBlockAlign;
	waveFmt.cbSize = (WORD)0;

fprintf(stderr,"sndout_init\n");
fflush(stderr);
	WaveBuf = GlobalAlloc(GPTR, NR_WAVEHDR * WAVEBUF_SIZE);
fprintf(stderr,"end GlobalAlloc %08x\n",WaveBuf);
fflush(stderr);
	memset(WaveHdr, 0, sizeof(WAVEHDR) * NR_WAVEHDR);
	for (i=0; i<NR_WAVEHDR; i++) {
		WaveHdr[i].lpData = WaveBuf+ i * WAVEBUF_SIZE;
		WaveHdr[i].dwBufferLength = WAVEBUF_SIZE;
	}

	r = waveOutOpen(&HWaveOut, WAVE_MAPPER, &waveFmt, 0, 0, CALLBACK_NULL);
	if (r != MMSYSERR_NOERROR) {
		fprintf(stderr,"waveOutOpen Error\n");
		exit(1);
	}
fprintf(stderr,"end waveOutOpen r = %d\n",r);
fflush(stderr);
	for (i=0; i<NR_WAVEHDR; i++) {
		waveOutPrepareHeader(HWaveOut,WaveHdr + i, sizeof(WAVEHDR));
	}
fprintf(stderr,"end waveOutPrepareHeader %08x \n",WaveHdr[0].dwFlags);
fflush(stderr);
}

static void sndout(int8_t d) {
	int next;

	next = cur_wavehdr + 1;
	if (next >= NR_WAVEHDR) next -= NR_WAVEHDR;
	if (cur_wavebuf_len >= WAVEBUF_SIZE) {
//fprintf(stderr,"start waveOutWrite\n");
//fflush(stderr);
		while (!(WaveHdr[next].dwFlags & WHDR_DONE)
		       && (WaveHdr[next].dwFlags != WHDR_PREPARED)) {
			fprintf(stderr,"wave buffer overflow\n");
			exit(1);
		}
//fprintf(stderr,"waveOutWrite #%d\n",cur_wavehdr);
//fflush(stderr);
		waveOutWrite(HWaveOut, WaveHdr + cur_wavehdr, sizeof(WAVEHDR));
		cur_wavehdr = next;
		cur_wavebuf_len = 0;
	}
	WaveHdr[cur_wavehdr].lpData[cur_wavebuf_len++] = d+128;
}

static int can_sndout() {
	int prev;

	prev = cur_wavehdr - 5;
	if (prev < 0) prev += NR_WAVEHDR;
	if  (WaveHdr[prev].dwFlags == WHDR_PREPARED) return 1;
	if  (WaveHdr[prev].dwFlags & WHDR_DONE) return 1;
	return 0;
}

#endif


prog_char sintab[64] = {
    0,    3,    6,    9,   12,   15,   18,   21,
   24,   27,   30,   33,   36,   39,   42,   45,
   48,   51,   54,   57,   59,   62,   65,   67,
   70,   73,   75,   78,   80,   82,   85,   87,
   89,   91,   94,   96,   98,  100,  102,  103,
  105,  107,  108,  110,  112,  113,  114,  116,
  117,  118,  119,  120,  121,  122,  123,  123,
  124,  125,  125,  126,  126,  126,  126,  126,
};

// warning: max RAMPLE_RATE = 16K
prog_uint16_t note_tab[12] = {
	137167254/ SAMPLE_RATE, // C7 = 108
	145323630/ SAMPLE_RATE,
	153965009/ SAMPLE_RATE,
	163120231/ SAMPLE_RATE,
	172819849/ SAMPLE_RATE,
	183096236/ SAMPLE_RATE,
	193983687/ SAMPLE_RATE,
	205518539/ SAMPLE_RATE,
	217739288/ SAMPLE_RATE,
	230686720/ SAMPLE_RATE, // A7 3520 Hz
	244404044/ SAMPLE_RATE,
	258937042/ SAMPLE_RATE,
};

struct LFO {
	uint16_t wave_pos;
	int16_t wave_inc;  // 2601 = 20Hz (20 * 255 * 16/SAMPLE_RATE)
	uint8_t intensity; // 255 = 6.225%
} lfo;

struct VCO {
	uint16_t wave_pos;
	int16_t wave_inc;
	int8_t lfo_inc;
	uint8_t wave_type;
} vco;

#define EG_NOTEOFF	0
#define EG_NOTEON	1
#define EG_ATTACK 	2
#define EG_DECAY	3
#define EG_SUSTAIN	4
#define EG_RELEASE	5

/* EG tick = 8000 / 16 = 500 Hz */
struct EG {
	uint8_t state;
	uint8_t out;
	uint16_t attack;
	uint16_t decay;
	uint16_t sustain_time; // 500Hz : max 131 sec
	uint16_t release;
	uint16_t sustain_lvl; // 0xffff : 100 %
	uint16_t sustain_cnt; // work
	uint16_t intval; // work
} eg;

static int16_t fmul16(uint16_t a, uint16_t b) {
	uint32_t v;
	v = a * b;
	return (v >> 16);
}

static int16_t fmulsu16(int16_t a, uint16_t b) {
	int32_t v;
	v = a * b;
	return (v >> 16);
}

static int16_t fmulsu16_8(int16_t a, uint8_t b) {
	int32_t v;
	v = a * b;
	return (v >> 8);
}

static int8_t fmulsu8_8(int8_t a, uint8_t b) {
	int16_t v;
	v = a * b;
	return (v >> 8);
}


static void note_on(uint8_t note) {
	uint16_t v;
	uint8_t shift = 0;
	if (note >= 120) return; 
	while (note < 108) {
		note += 12;
		shift++;
	}
	v = __LPM_word(note_tab+note-108);
	v >>= shift;
	vco.wave_inc = v;
	eg.state = EG_NOTEON;
}

static void do_eg() {
	uint8_t state = eg.state;
	if (state == EG_NOTEOFF) {
		eg.intval = (0 << 8);
	}
	if (state == EG_NOTEON) {
		eg.sustain_cnt = 0;
		if (eg.attack == 0) {
			eg.intval = eg.sustain_lvl;
			eg.state = state = EG_SUSTAIN;
fprintf(stderr,"EG_SUSTAIN\n");
		} else {
			eg.intval = 0;
			eg.state = state = EG_ATTACK;
fprintf(stderr,"EG_ATTACK %04x\n",eg.attack);
		}
	}
	if (state == EG_ATTACK) {
		if ((uint32_t)eg.intval + eg.attack >= (255<<8)) { // overflow
			eg.intval = (255 << 8);
			if (eg.sustain_lvl == 0) {
fprintf(stderr,"EG_RELEASE\n");
				eg.state = EG_RELEASE;
			} else {
fprintf(stderr,"EG_DECAY\n");
				eg.state = EG_DECAY;
			}
		} else {
			eg.intval += eg.attack;
		}
	} else if (state == EG_DECAY) {
		eg.intval = fmul16(eg.intval, eg.decay);
		if (eg.intval <= eg.sustain_lvl) {
			eg.intval = eg.sustain_lvl;
			if (eg.sustain_time == 0) {
fprintf(stderr,"EG_RELEASE\n");
				eg.state = EG_RELEASE;
			} else {
fprintf(stderr,"EG_SUSTAIN\n");
				eg.state = EG_SUSTAIN;
			}
		}
	} else if (state == EG_SUSTAIN) {
		eg.sustain_cnt ++;
		if (eg.sustain_cnt == eg.sustain_time) {
			eg.state = EG_RELEASE;
fprintf(stderr,"EG_RELEASE\n");
		}
	} else if (state == EG_RELEASE) {
		eg.intval = fmul16(eg.intval, eg.release);
		if (eg.intval < (1 << 8)) {
			eg.intval = 0;
			eg.state = EG_NOTEOFF;
fprintf(stderr,"EG_NOTEOFF\n");
		}
	}
	eg.out = eg.intval >> 8;
}

static void do_lfo() {
	uint8_t pos;
	int8_t rev = 0;
	int8_t v;
	int16_t v2;
	lfo.wave_pos += lfo.wave_inc;
	pos = lfo.wave_pos >> 8;
	if (pos >= 128) {
		rev = 1;
		pos = 255 - pos;
	}
	if (pos >= 64) {
		pos = 127 - pos;
	}
	v = __LPM(sintab+pos);
	if (rev) v = -v;
	v2 = (int16_t)v * lfo.intensity;
	vco.lfo_inc = fmulsu16(v2, vco.wave_inc)>>4;
}

static void generate_snd() {
	uint8_t pos;
	int8_t rev = 0;
	int8_t v;
	static uint8_t cnt;
	vco.wave_pos += vco.wave_inc + vco.lfo_inc;
	pos = vco.wave_pos >> 8;
	if (pos >= 128) {
		rev = 1;
		pos = 255 - pos;
	}
	if (vco.wave_type == WAVE_SAW) {
		v = 127 - pos;
	} else {
		if (pos >= 64) {
			pos = 127 - pos;
		}
		if (vco.wave_type == WAVE_TRI) {
			v = pos*2;
		} else 
			v = sintab[pos];
	}
	if (rev) v = -v;
	cnt++;
	if (cnt >= 16) {
		do_lfo();
		cnt = 0;
	} else if (cnt == 8) {
		do_eg();
	}
	v = fmulsu8_8(v, eg.out);
	sndout(v);
}

static void change_note(uint8_t c) {
   const uint8_t note_base = 36+24; // C1
   static int oct;
    switch(c) {
    case 'z': oct --; break;
    case 'x': oct ++; break;
    case 'q': exit(0);
    case 'w': vco.wave_type = WAVE_SIN; break; 
    case 'e': vco.wave_type = WAVE_TRI; break; 
    case 'r': vco.wave_type = WAVE_SAW; break; 
    case 'a': note_on(note_base + 12*oct +0); break;
    case 's': note_on(note_base + 12*oct +2); break;
    case 'd': note_on(note_base + 12*oct +4); break;
    case 'f': note_on(note_base + 12*oct +5); break;
    case 'g': note_on(note_base + 12*oct +7); break;
    case 'h': note_on(note_base + 12*oct +9); break;
    case 'j': note_on(note_base + 12*oct +11); break;
    case 'k': note_on(note_base + 12*oct +12); break;
    case 'l': note_on(note_base + 12*oct +14); break;
    }
}

main(int argc,char *argv[]) {
	uint8_t c;
	char *arg;
	sndout_init();
#ifdef LFO_TEST
	lfo.intensity = 255;
	lfo.wave_inc = 262; // 2 Hz
#endif
#if 0 // for demo
	eg.attack = 64;
	eg.sustain_lvl = (128<<8);
	eg.sustain_time = (12<<8);
	eg.decay = 0xfff0;
	eg.release = 0xff80;
#endif
#if 1
	eg.attack = 512*4;
	eg.sustain_lvl = (12<<8);
	eg.sustain_time = 0;
	eg.decay = 0xff80;
	eg.release = 0xf000;
#endif
#if 0
	eg.sustain_lvl = (255<<8);
	eg.sustain_time = 0xfff;
#endif
	note_on(36+24);

	if (argc > 1) {
		int hz;
		arg = argv[1];
		while (*arg >= 'a' && *arg <= 'z') {
			change_note(*arg++);
		}
	}
#if defined(__linux__) || defined(_WIN32)
	lin_init();
#endif
	for (;;) {
		if (can_sndout()) {
			generate_snd();
			continue;
		}
		if (usbcdc_can_getc(1)) {
			c = usbcdc_getc();
			change_note(c);
		}
	}
}
